import itertools
import math
from collections import defaultdict

from pint import (
    DefinitionSyntaxError,
    DimensionalityError,
    UndefinedUnitError,
    UnitRegistry,
)
from pint.context import Context
from pint.testsuite import QuantityTestCase
from pint.util import UnitsContainer


def add_ctxs(ureg):
    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[time]": -1})
    d = Context("lc")
    d.add_transformation(a, b, lambda ureg, x: ureg.speed_of_light / x)
    d.add_transformation(b, a, lambda ureg, x: ureg.speed_of_light / x)

    ureg.add_context(d)

    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[current]": 1})
    d = Context("ab")
    d.add_transformation(a, b, lambda ureg, x: ureg.ampere * ureg.meter / x)
    d.add_transformation(b, a, lambda ureg, x: ureg.ampere * ureg.meter / x)

    ureg.add_context(d)


def add_arg_ctxs(ureg):
    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[time]": -1})
    d = Context("lc")
    d.add_transformation(a, b, lambda ureg, x, n: ureg.speed_of_light / x / n)
    d.add_transformation(b, a, lambda ureg, x, n: ureg.speed_of_light / x / n)

    ureg.add_context(d)

    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[current]": 1})
    d = Context("ab")
    d.add_transformation(a, b, lambda ureg, x: ureg.ampere * ureg.meter / x)
    d.add_transformation(b, a, lambda ureg, x: ureg.ampere * ureg.meter / x)

    ureg.add_context(d)


def add_argdef_ctxs(ureg):
    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[time]": -1})
    d = Context("lc", defaults=dict(n=1))
    assert d.defaults == dict(n=1)

    d.add_transformation(a, b, lambda ureg, x, n: ureg.speed_of_light / x / n)
    d.add_transformation(b, a, lambda ureg, x, n: ureg.speed_of_light / x / n)

    ureg.add_context(d)

    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[current]": 1})
    d = Context("ab")
    d.add_transformation(a, b, lambda ureg, x: ureg.ampere * ureg.meter / x)
    d.add_transformation(b, a, lambda ureg, x: ureg.ampere * ureg.meter / x)

    ureg.add_context(d)


def add_sharedargdef_ctxs(ureg):
    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[time]": -1})
    d = Context("lc", defaults=dict(n=1))
    assert d.defaults == dict(n=1)

    d.add_transformation(a, b, lambda ureg, x, n: ureg.speed_of_light / x / n)
    d.add_transformation(b, a, lambda ureg, x, n: ureg.speed_of_light / x / n)

    ureg.add_context(d)

    a, b = UnitsContainer({"[length]": 1}), UnitsContainer({"[current]": 1})
    d = Context("ab", defaults=dict(n=0))
    d.add_transformation(a, b, lambda ureg, x, n: ureg.ampere * ureg.meter * n / x)
    d.add_transformation(b, a, lambda ureg, x, n: ureg.ampere * ureg.meter * n / x)

    ureg.add_context(d)


class TestContexts(QuantityTestCase):
    def test_known_context(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)
        with ureg.context("lc"):
            self.assertTrue(ureg._active_ctx)
            self.assertTrue(ureg._active_ctx.graph)

        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

        with ureg.context("lc", n=1):
            self.assertTrue(ureg._active_ctx)
            self.assertTrue(ureg._active_ctx.graph)

        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

    def test_known_context_enable(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)
        ureg.enable_contexts("lc")
        self.assertTrue(ureg._active_ctx)
        self.assertTrue(ureg._active_ctx.graph)
        ureg.disable_contexts(1)

        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

        ureg.enable_contexts("lc", n=1)
        self.assertTrue(ureg._active_ctx)
        self.assertTrue(ureg._active_ctx.graph)
        ureg.disable_contexts(1)

        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

    def test_graph(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)
        l = UnitsContainer({"[length]": 1.0})  # noqa: E741
        t = UnitsContainer({"[time]": -1.0})
        c = UnitsContainer({"[current]": 1.0})

        g_sp = defaultdict(set)
        g_sp.update({l: {t}, t: {l}})

        g_ab = defaultdict(set)
        g_ab.update({l: {c}, c: {l}})

        g = defaultdict(set)
        g.update({l: {t, c}, t: {l}, c: {l}})

        with ureg.context("lc"):
            self.assertEqual(ureg._active_ctx.graph, g_sp)

        with ureg.context("lc", n=1):
            self.assertEqual(ureg._active_ctx.graph, g_sp)

        with ureg.context("ab"):
            self.assertEqual(ureg._active_ctx.graph, g_ab)

        with ureg.context("lc"):
            with ureg.context("ab"):
                self.assertEqual(ureg._active_ctx.graph, g)

        with ureg.context("ab"):
            with ureg.context("lc"):
                self.assertEqual(ureg._active_ctx.graph, g)

        with ureg.context("lc", "ab"):
            self.assertEqual(ureg._active_ctx.graph, g)

        with ureg.context("ab", "lc"):
            self.assertEqual(ureg._active_ctx.graph, g)

    def test_graph_enable(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)
        l = UnitsContainer({"[length]": 1.0})  # noqa: E741
        t = UnitsContainer({"[time]": -1.0})
        c = UnitsContainer({"[current]": 1.0})

        g_sp = defaultdict(set)
        g_sp.update({l: {t}, t: {l}})

        g_ab = defaultdict(set)
        g_ab.update({l: {c}, c: {l}})

        g = defaultdict(set)
        g.update({l: {t, c}, t: {l}, c: {l}})

        ureg.enable_contexts("lc")
        self.assertEqual(ureg._active_ctx.graph, g_sp)
        ureg.disable_contexts(1)

        ureg.enable_contexts("lc", n=1)
        self.assertEqual(ureg._active_ctx.graph, g_sp)
        ureg.disable_contexts(1)

        ureg.enable_contexts("ab")
        self.assertEqual(ureg._active_ctx.graph, g_ab)
        ureg.disable_contexts(1)

        ureg.enable_contexts("lc")
        ureg.enable_contexts("ab")
        self.assertEqual(ureg._active_ctx.graph, g)
        ureg.disable_contexts(2)

        ureg.enable_contexts("ab")
        ureg.enable_contexts("lc")
        self.assertEqual(ureg._active_ctx.graph, g)
        ureg.disable_contexts(2)

        ureg.enable_contexts("lc", "ab")
        self.assertEqual(ureg._active_ctx.graph, g)
        ureg.disable_contexts(2)

        ureg.enable_contexts("ab", "lc")
        self.assertEqual(ureg._active_ctx.graph, g)
        ureg.disable_contexts(2)

    def test_known_nested_context(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)

        with ureg.context("lc"):
            x = dict(ureg._active_ctx)
            y = dict(ureg._active_ctx.graph)
            self.assertTrue(ureg._active_ctx)
            self.assertTrue(ureg._active_ctx.graph)

            with ureg.context("ab"):
                self.assertTrue(ureg._active_ctx)
                self.assertTrue(ureg._active_ctx.graph)
                self.assertNotEqual(x, ureg._active_ctx)
                self.assertNotEqual(y, ureg._active_ctx.graph)

            self.assertEqual(x, ureg._active_ctx)
            self.assertEqual(y, ureg._active_ctx.graph)

        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

    def test_unknown_context(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)
        with self.assertRaises(KeyError):
            with ureg.context("la"):
                pass
        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

    def test_unknown_nested_context(self):
        ureg = UnitRegistry()
        add_ctxs(ureg)

        with ureg.context("lc"):
            x = dict(ureg._active_ctx)
            y = dict(ureg._active_ctx.graph)
            with self.assertRaises(KeyError):
                with ureg.context("la"):
                    pass

            self.assertEqual(x, ureg._active_ctx)
            self.assertEqual(y, ureg._active_ctx.graph)

        self.assertFalse(ureg._active_ctx)
        self.assertFalse(ureg._active_ctx.graph)

    def test_one_context(self):
        ureg = UnitRegistry()

        add_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        meter_units = ureg.get_compatible_units(ureg.meter)
        hertz_units = ureg.get_compatible_units(ureg.hertz)

        self.assertRaises(DimensionalityError, q.to, "Hz")
        with ureg.context("lc"):
            self.assertEqual(q.to("Hz"), s)
            self.assertEqual(ureg.get_compatible_units(q), meter_units | hertz_units)
        self.assertRaises(DimensionalityError, q.to, "Hz")
        self.assertEqual(ureg.get_compatible_units(q), meter_units)

    def test_multiple_context(self):
        ureg = UnitRegistry()

        add_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        meter_units = ureg.get_compatible_units(ureg.meter)
        hertz_units = ureg.get_compatible_units(ureg.hertz)
        ampere_units = ureg.get_compatible_units(ureg.ampere)

        self.assertRaises(DimensionalityError, q.to, "Hz")
        with ureg.context("lc", "ab"):
            self.assertEqual(q.to("Hz"), s)
            self.assertEqual(
                ureg.get_compatible_units(q), meter_units | hertz_units | ampere_units
            )
        self.assertRaises(DimensionalityError, q.to, "Hz")
        self.assertEqual(ureg.get_compatible_units(q), meter_units)

    def test_nested_context(self):
        ureg = UnitRegistry()

        add_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        self.assertRaises(DimensionalityError, q.to, "Hz")
        with ureg.context("lc"):
            self.assertEqual(q.to("Hz"), s)
            with ureg.context("ab"):
                self.assertEqual(q.to("Hz"), s)
            self.assertEqual(q.to("Hz"), s)

        with ureg.context("ab"):
            self.assertRaises(DimensionalityError, q.to, "Hz")
            with ureg.context("lc"):
                self.assertEqual(q.to("Hz"), s)
            self.assertRaises(DimensionalityError, q.to, "Hz")

    def test_context_with_arg(self):

        ureg = UnitRegistry()

        add_arg_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        self.assertRaises(DimensionalityError, q.to, "Hz")
        with ureg.context("lc", n=1):
            self.assertEqual(q.to("Hz"), s)
            with ureg.context("ab"):
                self.assertEqual(q.to("Hz"), s)
            self.assertEqual(q.to("Hz"), s)

        with ureg.context("ab"):
            self.assertRaises(DimensionalityError, q.to, "Hz")
            with ureg.context("lc", n=1):
                self.assertEqual(q.to("Hz"), s)
            self.assertRaises(DimensionalityError, q.to, "Hz")

        with ureg.context("lc"):
            self.assertRaises(TypeError, q.to, "Hz")

    def test_enable_context_with_arg(self):

        ureg = UnitRegistry()

        add_arg_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        self.assertRaises(DimensionalityError, q.to, "Hz")
        ureg.enable_contexts("lc", n=1)
        self.assertEqual(q.to("Hz"), s)
        ureg.enable_contexts("ab")
        self.assertEqual(q.to("Hz"), s)
        self.assertEqual(q.to("Hz"), s)
        ureg.disable_contexts(1)
        ureg.disable_contexts(1)

        ureg.enable_contexts("ab")
        self.assertRaises(DimensionalityError, q.to, "Hz")
        ureg.enable_contexts("lc", n=1)
        self.assertEqual(q.to("Hz"), s)
        ureg.disable_contexts(1)
        self.assertRaises(DimensionalityError, q.to, "Hz")
        ureg.disable_contexts(1)

        ureg.enable_contexts("lc")
        self.assertRaises(TypeError, q.to, "Hz")
        ureg.disable_contexts(1)

    def test_context_with_arg_def(self):

        ureg = UnitRegistry()

        add_argdef_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        self.assertRaises(DimensionalityError, q.to, "Hz")
        with ureg.context("lc"):
            self.assertEqual(q.to("Hz"), s)
            with ureg.context("ab"):
                self.assertEqual(q.to("Hz"), s)
            self.assertEqual(q.to("Hz"), s)

        with ureg.context("ab"):
            self.assertRaises(DimensionalityError, q.to, "Hz")
            with ureg.context("lc"):
                self.assertEqual(q.to("Hz"), s)
            self.assertRaises(DimensionalityError, q.to, "Hz")

        self.assertRaises(DimensionalityError, q.to, "Hz")
        with ureg.context("lc", n=2):
            self.assertEqual(q.to("Hz"), s / 2)
            with ureg.context("ab"):
                self.assertEqual(q.to("Hz"), s / 2)
            self.assertEqual(q.to("Hz"), s / 2)

        with ureg.context("ab"):
            self.assertRaises(DimensionalityError, q.to, "Hz")
            with ureg.context("lc", n=2):
                self.assertEqual(q.to("Hz"), s / 2)
            self.assertRaises(DimensionalityError, q.to, "Hz")

    def test_context_with_sharedarg_def(self):

        ureg = UnitRegistry()

        add_sharedargdef_ctxs(ureg)

        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")
        u = (1 / 500) * ureg.ampere

        with ureg.context("lc"):
            self.assertEqual(q.to("Hz"), s)
            with ureg.context("ab"):
                self.assertEqual(q.to("ampere"), u)

        with ureg.context("ab"):
            self.assertEqual(q.to("ampere"), 0 * u)
            with ureg.context("lc"):
                self.assertRaises(ZeroDivisionError, ureg.Quantity.to, q, "Hz")

        with ureg.context("lc", n=2):
            self.assertEqual(q.to("Hz"), s / 2)
            with ureg.context("ab"):
                self.assertEqual(q.to("ampere"), 2 * u)

        with ureg.context("ab", n=3):
            self.assertEqual(q.to("ampere"), 3 * u)
            with ureg.context("lc"):
                self.assertEqual(q.to("Hz"), s / 3)

        with ureg.context("lc", n=2):
            self.assertEqual(q.to("Hz"), s / 2)
            with ureg.context("ab", n=4):
                self.assertEqual(q.to("ampere"), 4 * u)

        with ureg.context("ab", n=3):
            self.assertEqual(q.to("ampere"), 3 * u)
            with ureg.context("lc", n=6):
                self.assertEqual(q.to("Hz"), s / 6)

    def test_anonymous_context(self):
        ureg = UnitRegistry()
        c = Context()
        c.add_transformation("[length]", "[time]", lambda ureg, x: x / ureg("5 cm/s"))
        self.assertRaises(ValueError, ureg.add_context, c)

        x = ureg("10 cm")
        expect = ureg("2 s")
        self.assertQuantityEqual(x.to("s", c), expect)

        with ureg.context(c):
            self.assertQuantityEqual(x.to("s"), expect)

        ureg.enable_contexts(c)
        self.assertQuantityEqual(x.to("s"), expect)
        ureg.disable_contexts(1)
        self.assertRaises(DimensionalityError, x.to, "s")

        # Multiple anonymous contexts
        c2 = Context()
        c2.add_transformation("[length]", "[time]", lambda ureg, x: x / ureg("10 cm/s"))
        c2.add_transformation("[mass]", "[time]", lambda ureg, x: x / ureg("10 kg/s"))
        with ureg.context(c2, c):
            self.assertQuantityEqual(x.to("s"), expect)
            # Transformations only in c2 are still working even if c takes priority
            self.assertQuantityEqual(ureg("100 kg").to("s"), ureg("10 s"))
        with ureg.context(c, c2):
            self.assertQuantityEqual(x.to("s"), ureg("1 s"))

    def _test_ctx(self, ctx):
        ureg = UnitRegistry()
        q = 500 * ureg.meter
        s = (ureg.speed_of_light / q).to("Hz")

        nctx = len(ureg._contexts)

        self.assertNotIn(ctx.name, ureg._contexts)
        ureg.add_context(ctx)

        self.assertIn(ctx.name, ureg._contexts)
        self.assertEqual(len(ureg._contexts), nctx + 1 + len(ctx.aliases))

        with ureg.context(ctx.name):
            self.assertEqual(q.to("Hz"), s)
            self.assertEqual(s.to("meter"), q)

        ureg.remove_context(ctx.name)
        self.assertNotIn(ctx.name, ureg._contexts)
        self.assertEqual(len(ureg._contexts), nctx)

    def test_parse_invalid(self):
        for badrow in (
            "[length] = 1 / [time]: c / value",
            "1 / [time] = [length]: c / value",
            "[length] <- [time] = c / value",
            "[length] - [time] = c / value",
        ):
            with self.subTest(badrow):
                with self.assertRaises(DefinitionSyntaxError):
                    Context.from_lines(["@context c", badrow])

    def test_parse_simple(self):

        a = Context.__keytransform__(
            UnitsContainer({"[time]": -1}), UnitsContainer({"[length]": 1})
        )
        b = Context.__keytransform__(
            UnitsContainer({"[length]": 1}), UnitsContainer({"[time]": -1})
        )

        s = [
            "@context longcontextname",
            "[length] -> 1 / [time]: c / value",
            "1 / [time] -> [length]: c / value",
        ]

        c = Context.from_lines(s)
        self.assertEqual(c.name, "longcontextname")
        self.assertEqual(c.aliases, ())
        self.assertEqual(c.defaults, {})
        self.assertEqual(c.funcs.keys(), {a, b})
        self._test_ctx(c)

        s = ["@context longcontextname = lc", "[length] <-> 1 / [time]: c / value"]

        c = Context.from_lines(s)
        self.assertEqual(c.name, "longcontextname")
        self.assertEqual(c.aliases, ("lc",))
        self.assertEqual(c.defaults, {})
        self.assertEqual(c.funcs.keys(), {a, b})
        self._test_ctx(c)

        s = [
            "@context longcontextname = lc = lcn",
            "[length] <-> 1 / [time]: c / value",
        ]

        c = Context.from_lines(s)
        self.assertEqual(c.name, "longcontextname")
        self.assertEqual(c.aliases, ("lc", "lcn"))
        self.assertEqual(c.defaults, {})
        self.assertEqual(c.funcs.keys(), {a, b})
        self._test_ctx(c)

    def test_parse_auto_inverse(self):

        a = Context.__keytransform__(
            UnitsContainer({"[time]": -1.0}), UnitsContainer({"[length]": 1.0})
        )
        b = Context.__keytransform__(
            UnitsContainer({"[length]": 1.0}), UnitsContainer({"[time]": -1.0})
        )

        s = ["@context longcontextname", "[length] <-> 1 / [time]: c / value"]

        c = Context.from_lines(s)
        self.assertEqual(c.defaults, {})
        self.assertEqual(c.funcs.keys(), {a, b})
        self._test_ctx(c)

    def test_parse_define(self):
        a = Context.__keytransform__(
            UnitsContainer({"[time]": -1}), UnitsContainer({"[length]": 1.0})
        )
        b = Context.__keytransform__(
            UnitsContainer({"[length]": 1}), UnitsContainer({"[time]": -1.0})
        )

        s = ["@context longcontextname", "[length] <-> 1 / [time]: c / value"]
        c = Context.from_lines(s)
        self.assertEqual(c.defaults, {})
        self.assertEqual(c.funcs.keys(), {a, b})
        self._test_ctx(c)

    def test_parse_parameterized(self):
        a = Context.__keytransform__(
            UnitsContainer({"[time]": -1.0}), UnitsContainer({"[length]": 1.0})
        )
        b = Context.__keytransform__(
            UnitsContainer({"[length]": 1.0}), UnitsContainer({"[time]": -1.0})
        )

        s = ["@context(n=1) longcontextname", "[length] <-> 1 / [time]: n * c / value"]

        c = Context.from_lines(s)
        self.assertEqual(c.defaults, {"n": 1})
        self.assertEqual(c.funcs.keys(), {a, b})
        self._test_ctx(c)

        s = [
            "@context(n=1, bla=2) longcontextname",
            "[length] <-> 1 / [time]: n * c / value / bla",
        ]

        c = Context.from_lines(s)
        self.assertEqual(c.defaults, {"n": 1, "bla": 2})
        self.assertEqual(c.funcs.keys(), {a, b})

        # If the variable is not present in the definition, then raise an error
        s = ["@context(n=1) longcontextname", "[length] <-> 1 / [time]: c / value"]
        self.assertRaises(DefinitionSyntaxError, Context.from_lines, s)

    def test_warnings(self):

        ureg = UnitRegistry()

        with self.capture_log() as buffer:
            add_ctxs(ureg)

            d = Context("ab")
            ureg.add_context(d)

            self.assertEqual(len(buffer), 1)
            self.assertIn("ab", str(buffer[-1]))

            d = Context("ab1", aliases=("ab",))
            ureg.add_context(d)

            self.assertEqual(len(buffer), 2)
            self.assertIn("ab", str(buffer[-1]))


class TestDefinedContexts(QuantityTestCase):

    FORCE_NDARRAY = False

    def test_defined(self):
        ureg = self.ureg
        with ureg.context("sp"):
            pass

        a = Context.__keytransform__(
            UnitsContainer({"[time]": -1.0}), UnitsContainer({"[length]": 1.0})
        )
        b = Context.__keytransform__(
            UnitsContainer({"[length]": 1.0}), UnitsContainer({"[time]": -1.0})
        )
        self.assertIn(a, ureg._contexts["sp"].funcs)
        self.assertIn(b, ureg._contexts["sp"].funcs)
        with ureg.context("sp"):
            self.assertIn(a, ureg._active_ctx)
            self.assertIn(b, ureg._active_ctx)

    def test_spectroscopy(self):
        ureg = self.ureg
        eq = (532.0 * ureg.nm, 563.5 * ureg.terahertz, 2.33053 * ureg.eV)
        with ureg.context("sp"):
            from pint.util import find_shortest_path

            for a, b in itertools.product(eq, eq):
                for x in range(2):
                    if x == 1:
                        a = a.to_base_units()
                        b = b.to_base_units()
                    da, db = Context.__keytransform__(
                        a.dimensionality, b.dimensionality
                    )
                    p = find_shortest_path(ureg._active_ctx.graph, da, db)
                    self.assertTrue(p)
                    msg = "{} <-> {}".format(a, b)
                    # assertAlmostEqualRelError converts second to first
                    self.assertQuantityAlmostEqual(b, a, rtol=0.01, msg=msg)

        for a, b in itertools.product(eq, eq):
            self.assertQuantityAlmostEqual(a.to(b.units, "sp"), b, rtol=0.01)

    def test_textile(self):
        ureg = self.ureg
        qty_direct = 1.331 * ureg.tex
        with self.assertRaises(DimensionalityError):
            qty_indirect = qty_direct.to("Nm")

        with ureg.context("textile"):
            from pint.util import find_shortest_path

            qty_indirect = qty_direct.to("Nm")
            a = qty_direct.to_base_units()
            b = qty_indirect.to_base_units()
            da, db = Context.__keytransform__(a.dimensionality, b.dimensionality)
            p = find_shortest_path(ureg._active_ctx.graph, da, db)
            self.assertTrue(p)
            msg = "{} <-> {}".format(a, b)
            self.assertQuantityAlmostEqual(b, a, rtol=0.01, msg=msg)

            # Check RKM <-> cN/tex conversion
            self.assertQuantityAlmostEqual(1 * ureg.RKM, 0.980665 * ureg.cN / ureg.tex)
            self.assertQuantityAlmostEqual(
                (1 / 0.980665) * ureg.RKM, 1 * ureg.cN / ureg.tex
            )
            self.assertAlmostEqual((1 * ureg.RKM).to(ureg.cN / ureg.tex).m, 0.980665)
            self.assertAlmostEqual(
                (1 * ureg.cN / ureg.tex).to(ureg.RKM).m, 1 / 0.980665
            )

    def test_decorator(self):
        ureg = self.ureg

        a = 532.0 * ureg.nm
        with ureg.context("sp"):
            b = a.to("terahertz")

        def f(wl):
            return wl.to("terahertz")

        self.assertRaises(DimensionalityError, f, a)

        @ureg.with_context("sp")
        def g(wl):
            return wl.to("terahertz")

        self.assertEqual(b, g(a))

    def test_decorator_composition(self):
        ureg = self.ureg

        a = 532.0 * ureg.nm
        with ureg.context("sp"):
            b = a.to("terahertz")

        @ureg.with_context("sp")
        @ureg.check("[length]")
        def f(wl):
            return wl.to("terahertz")

        @ureg.with_context("sp")
        @ureg.check("[length]")
        def g(wl):
            return wl.to("terahertz")

        self.assertEqual(b, f(a))
        self.assertEqual(b, g(a))


class TestContextRedefinitions(QuantityTestCase):
    def test_redefine(self):
        ureg = UnitRegistry(
            """
            foo = [d] = f = foo_alias
            bar = 2 foo = b = bar_alias
            baz = 3 bar = _ = baz_alias
            asd = 4 baz

            @context c
                # Note how we're redefining a symbol, not the base name, as a
                # function of another name
                b = 5 f
            """.splitlines()
        )
        # Units that are somehow directly or indirectly defined as a function of the
        # overridden unit are also affected
        foo = ureg.Quantity(1, "foo")
        bar = ureg.Quantity(1, "bar")
        asd = ureg.Quantity(1, "asd")

        # Test without context before and after, to verify that the cache and units have
        # not been polluted
        for enable_ctx in (False, True, False):
            with self.subTest(enable_ctx):
                if enable_ctx:
                    ureg.enable_contexts("c")
                    k = 5
                else:
                    k = 2

                self.assertEqual(foo.to("b").magnitude, 1 / k)
                self.assertEqual(foo.to("bar").magnitude, 1 / k)
                self.assertEqual(foo.to("bar_alias").magnitude, 1 / k)
                self.assertEqual(foo.to("baz").magnitude, 1 / k / 3)
                self.assertEqual(bar.to("foo").magnitude, k)
                self.assertEqual(bar.to("baz").magnitude, 1 / 3)
                self.assertEqual(asd.to("foo").magnitude, 4 * 3 * k)
                self.assertEqual(asd.to("bar").magnitude, 4 * 3)
                self.assertEqual(asd.to("baz").magnitude, 4)

            ureg.disable_contexts()

    def test_define_nan(self):
        ureg = UnitRegistry(
            """
            USD = [currency]
            EUR = nan USD
            GBP = nan USD

            @context c
                EUR = 1.11 USD
                # Note that we're changing which unit GBP is defined against
                GBP = 1.18 EUR
            @end
            """.splitlines()
        )

        q = ureg.Quantity("10 GBP")
        self.assertEquals(q.magnitude, 10)
        self.assertEquals(q.units.dimensionality, {"[currency]": 1})
        self.assertEquals(q.to("GBP").magnitude, 10)
        self.assertTrue(math.isnan(q.to("USD").magnitude))
        self.assertAlmostEqual(q.to("USD", "c").magnitude, 10 * 1.18 * 1.11)

    def test_non_multiplicative(self):
        ureg = UnitRegistry(
            """
            kelvin = [temperature]
            fahrenheit = 5 / 9 * kelvin; offset: 255
            bogodegrees = 9 * kelvin

            @context nonmult_to_nonmult
                fahrenheit = 7 * kelvin; offset: 123
            @end
            @context nonmult_to_mult
                fahrenheit = 123 * kelvin
            @end
            @context mult_to_nonmult
                bogodegrees = 5 * kelvin; offset: 123
            @end
            """.splitlines()
        )
        k = ureg.Quantity(100, "kelvin")

        with self.subTest("baseline"):
            self.assertAlmostEqual(k.to("fahrenheit").magnitude, (100 - 255) * 9 / 5)
            self.assertAlmostEqual(k.to("bogodegrees").magnitude, 100 / 9)

        with self.subTest("nonmult_to_nonmult"):
            with ureg.context("nonmult_to_nonmult"):
                self.assertAlmostEqual(k.to("fahrenheit").magnitude, (100 - 123) / 7)

        with self.subTest("nonmult_to_mult"):
            with ureg.context("nonmult_to_mult"):
                self.assertAlmostEqual(k.to("fahrenheit").magnitude, 100 / 123)

        with self.subTest("mult_to_nonmult"):
            with ureg.context("mult_to_nonmult"):
                self.assertAlmostEqual(k.to("bogodegrees").magnitude, (100 - 123) / 5)

    def test_err_to_base_unit(self):
        with self.assertRaises(DefinitionSyntaxError) as e:
            Context.from_lines(["@context c", "x = [d]"])
        self.assertEquals(str(e.exception), "Can't define base units within a context")

    def test_err_change_base_unit(self):
        ureg = UnitRegistry(
            """
            foo = [d1]
            bar = [d2]

            @context c
                bar = foo
            @end
            """.splitlines()
        )

        with self.assertRaises(ValueError) as e:
            ureg.enable_contexts("c")
        self.assertEquals(
            str(e.exception), "Can't redefine a base unit to a derived one"
        )

    def test_err_change_dimensionality(self):
        ureg = UnitRegistry(
            """
            foo = [d1]
            bar = [d2]
            baz = foo

            @context c
                baz = bar
            @end
            """.splitlines()
        )
        with self.assertRaises(ValueError) as e:
            ureg.enable_contexts("c")
        self.assertEquals(
            str(e.exception),
            "Can't change dimensionality of baz from [d1] to [d2] in a context",
        )

    def test_err_cyclic_dependency(self):
        ureg = UnitRegistry(
            """
            foo = [d]
            bar = foo
            baz = bar

            @context c
                bar = baz
            @end
            """.splitlines()
        )
        # TODO align this exception and the one you get when you implement a cyclic
        #      dependency within the base registry. Ideally this exception should be
        #      raised by enable_contexts.
        ureg.enable_contexts("c")
        q = ureg.Quantity("bar")
        with self.assertRaises(RecursionError):
            q.to("foo")

    def test_err_dimension_redefinition(self):
        with self.assertRaises(DefinitionSyntaxError) as e:
            Context.from_lines(["@context c", "[d1] = [d2] * [d3]"])
        self.assertEquals(
            str(e.exception), "Expected <unit> = <converter>; got [d1] = [d2] * [d3]"
        )

    def test_err_prefix_redefinition(self):
        with self.assertRaises(DefinitionSyntaxError) as e:
            Context.from_lines(["@context c", "[d1] = [d2] * [d3]"])
        self.assertEquals(
            str(e.exception), "Expected <unit> = <converter>; got [d1] = [d2] * [d3]"
        )

    def test_err_redefine_alias(self):
        for s in ("foo = bar = f", "foo = bar = _ = baz"):
            with self.subTest(s):
                with self.assertRaises(DefinitionSyntaxError) as e:
                    Context.from_lines(["@context c", s])
                self.assertEquals(
                    str(e.exception),
                    "Can't change a unit's symbol or aliases within a context",
                )

    def test_err_redefine_with_prefix(self):
        ureg = UnitRegistry(
            """
            kilo- = 1000
            gram = [mass]
            pound = 454 gram

            @context c
                kilopound = 500000 gram
            @end
            """.splitlines()
        )
        with self.assertRaises(ValueError) as e:
            ureg.enable_contexts("c")
        self.assertEquals(
            str(e.exception), "Can't redefine a unit with a prefix: kilopound"
        )

    def test_err_new_unit(self):
        ureg = UnitRegistry(
            """
            foo = [d]
            @context c
                bar = foo
            @end
            """.splitlines()
        )
        with self.assertRaises(UndefinedUnitError) as e:
            ureg.enable_contexts("c")
        self.assertEquals(str(e.exception), "'bar' is not defined in the unit registry")



import copy
import functools
import math
import re

from pint import (
    DefinitionSyntaxError,
    DimensionalityError,
    RedefinitionError,
    UndefinedUnitError,
)
from pint.compat import np
from pint.registry import LazyRegistry, UnitRegistry
from pint.testsuite import QuantityTestCase, helpers
from pint.testsuite.parameterized import ParameterizedTestCase
from pint.util import ParserHelper, UnitsContainer


class TestUnit(QuantityTestCase):
    def test_creation(self):
        for arg in ("meter", UnitsContainer(meter=1), self.U_("m")):
            self.assertEqual(self.U_(arg)._units, UnitsContainer(meter=1))
        self.assertRaises(TypeError, self.U_, 1)

    def test_deepcopy(self):
        x = self.U_(UnitsContainer(meter=1))
        self.assertEqual(x, copy.deepcopy(x))

    def test_unit_repr(self):
        x = self.U_(UnitsContainer(meter=1))
        self.assertEqual(str(x), "meter")
        self.assertEqual(repr(x), "<Unit('meter')>")

    def test_unit_formatting(self):
        x = self.U_(UnitsContainer(meter=2, kilogram=1, second=-1))
        for spec, result in (
            ("{}", str(x)),
            ("{!s}", str(x)),
            ("{!r}", repr(x)),
            (
                "{:L}",
                r"\frac{\mathrm{kilogram} \cdot \mathrm{meter}^{2}}{\mathrm{second}}",
            ),
            ("{:P}", "kilogram·meter²/second"),
            ("{:H}", r"\[kilogram\ meter^2/second\]"),
            ("{:C}", "kilogram*meter**2/second"),
            ("{:Lx}", r"\si[]{\kilo\gram\meter\squared\per\second}"),
            ("{:~}", "kg * m ** 2 / s"),
            ("{:L~}", r"\frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}"),
            ("{:P~}", "kg·m²/s"),
            ("{:H~}", r"\[kg\ m^2/s\]"),
            ("{:C~}", "kg*m**2/s"),
        ):
            with self.subTest(spec):
                self.assertEqual(spec.format(x), result)

    def test_unit_default_formatting(self):
        ureg = UnitRegistry()
        x = ureg.Unit(UnitsContainer(meter=2, kilogram=1, second=-1))
        for spec, result in (
            (
                "L",
                r"\frac{\mathrm{kilogram} \cdot \mathrm{meter}^{2}}{\mathrm{second}}",
            ),
            ("P", "kilogram·meter²/second"),
            ("H", r"\[kilogram\ meter^2/second\]"),
            ("C", "kilogram*meter**2/second"),
            ("~", "kg * m ** 2 / s"),
            ("L~", r"\frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}"),
            ("P~", "kg·m²/s"),
            ("H~", r"\[kg\ m^2/s\]"),
            ("C~", "kg*m**2/s"),
        ):
            with self.subTest(spec):
                ureg.default_format = spec
                self.assertEqual(f"{x}", result, f"Failed for {spec}, {result}")

    def test_unit_formatting_snake_case(self):
        # Test that snake_case units are escaped where appropriate
        ureg = UnitRegistry()
        x = ureg.Unit(UnitsContainer(oil_barrel=1))
        for spec, result in (
            ("L", r"\mathrm{oil\_barrel}"),
            ("P", "oil_barrel"),
            ("H", r"\[oil\_barrel\]"),
            ("C", "oil_barrel"),
            ("~", "oil_bbl"),
            ("L~", r"\mathrm{oil\_bbl}"),
            ("P~", "oil_bbl"),
            ("H~", r"\[oil\_bbl\]"),
            ("C~", "oil_bbl"),
        ):
            with self.subTest(spec):
                ureg.default_format = spec
                self.assertEqual(f"{x}", result, f"Failed for {spec}, {result}")

    def test_ipython(self):
        alltext = []

        class Pretty:
            @staticmethod
            def text(text):
                alltext.append(text)

        ureg = UnitRegistry()
        x = ureg.Unit(UnitsContainer(meter=2, kilogram=1, second=-1))
        self.assertEqual(x._repr_html_(), r"\[kilogram\ meter^2/second\]")
        self.assertEqual(
            x._repr_latex_(),
            r"$\frac{\mathrm{kilogram} \cdot " r"\mathrm{meter}^{2}}{\mathrm{second}}$",
        )
        x._repr_pretty_(Pretty, False)
        self.assertEqual("".join(alltext), "kilogram·meter²/second")
        ureg.default_format = "~"
        self.assertEqual(x._repr_html_(), r"\[kg\ m^2/s\]")
        self.assertEqual(
            x._repr_latex_(), r"$\frac{\mathrm{kg} \cdot \mathrm{m}^{2}}{\mathrm{s}}$"
        )
        alltext = []
        x._repr_pretty_(Pretty, False)
        self.assertEqual("".join(alltext), "kg·m²/s")

    def test_unit_mul(self):
        x = self.U_("m")
        self.assertEqual(x * 1, self.Q_(1, "m"))
        self.assertEqual(x * 0.5, self.Q_(0.5, "m"))
        self.assertEqual(x * self.Q_(1, "m"), self.Q_(1, "m**2"))
        self.assertEqual(1 * x, self.Q_(1, "m"))

    def test_unit_div(self):
        x = self.U_("m")
        self.assertEqual(x / 1, self.Q_(1, "m"))
        self.assertEqual(x / 0.5, self.Q_(2.0, "m"))
        self.assertEqual(x / self.Q_(1, "m"), self.Q_(1))

    def test_unit_rdiv(self):
        x = self.U_("m")
        self.assertEqual(1 / x, self.Q_(1, "1/m"))

    def test_unit_pow(self):
        x = self.U_("m")
        self.assertEqual(x ** 2, self.U_("m**2"))

    def test_unit_hash(self):
        x = self.U_("m")
        self.assertEqual(hash(x), hash(x._units))

    def test_unit_eqs(self):
        x = self.U_("m")
        self.assertEqual(x, self.U_("m"))
        self.assertNotEqual(x, self.U_("cm"))

        self.assertEqual(x, self.Q_(1, "m"))
        self.assertNotEqual(x, self.Q_(2, "m"))

        self.assertEqual(x, UnitsContainer({"meter": 1}))

        y = self.U_("cm/m")
        self.assertEqual(y, 0.01)

        self.assertEqual(self.U_("byte") == self.U_("byte"), True)
        self.assertEqual(self.U_("byte") != self.U_("byte"), False)

    def test_unit_cmp(self):

        x = self.U_("m")
        self.assertLess(x, self.U_("km"))
        self.assertGreater(x, self.U_("mm"))

        y = self.U_("m/mm")
        self.assertGreater(y, 1)
        self.assertLess(y, 1e6)

    def test_dimensionality(self):

        x = self.U_("m")
        self.assertEqual(x.dimensionality, UnitsContainer({"[length]": 1}))

    def test_dimensionless(self):

        self.assertTrue(self.U_("m/mm").dimensionless)
        self.assertFalse(self.U_("m").dimensionless)

    def test_unit_casting(self):

        self.assertEqual(int(self.U_("m/mm")), 1000)
        self.assertEqual(float(self.U_("mm/m")), 1e-3)
        self.assertEqual(complex(self.U_("mm/mm")), 1 + 0j)

    @helpers.requires_numpy()
    def test_array_interface(self):
        import numpy as np

        x = self.U_("m")
        arr = np.ones(10)
        self.assertQuantityEqual(arr * x, self.Q_(arr, "m"))
        self.assertQuantityEqual(arr / x, self.Q_(arr, "1/m"))
        self.assertQuantityEqual(x / arr, self.Q_(arr, "m"))


class TestRegistry(QuantityTestCase):

    FORCE_NDARRAY = False

    def setup(self):
        self.ureg.autoconvert_offset_to_baseunit = False

    def test_base(self):
        ureg = UnitRegistry(None)
        ureg.define("meter = [length]")
        self.assertRaises(DefinitionSyntaxError, ureg.define, "meter = [length]")
        self.assertRaises(TypeError, ureg.define, list())
        ureg.define("degC = kelvin; offset: 273.15")

    def test_define(self):
        ureg = UnitRegistry(None)
        self.assertIsInstance(dir(ureg), list)
        self.assertGreater(len(dir(ureg)), 0)

    def test_load(self):
        import pkg_resources
        from pint import unit

        data = pkg_resources.resource_filename(unit.__name__, "default_en.txt")
        ureg1 = UnitRegistry()
        ureg2 = UnitRegistry(data)
        self.assertEqual(dir(ureg1), dir(ureg2))
        self.assertRaises(
            ValueError, UnitRegistry(None).load_definitions, "notexisting"
        )

    def test_default_format(self):
        ureg = UnitRegistry()
        q = ureg.meter
        s1 = f"{q}"
        s2 = f"{q:~}"
        ureg.default_format = "~"
        s3 = f"{q}"
        self.assertEqual(s2, s3)
        self.assertNotEqual(s1, s3)
        self.assertEqual(ureg.default_format, "~")

    def test_parse_number(self):
        self.assertEqual(self.ureg.parse_expression("pi"), math.pi)
        self.assertEqual(self.ureg.parse_expression("x", x=2), 2)
        self.assertEqual(self.ureg.parse_expression("x", x=2.3), 2.3)
        self.assertEqual(self.ureg.parse_expression("x * y", x=2.3, y=3), 2.3 * 3)
        self.assertEqual(self.ureg.parse_expression("x", x=(1 + 1j)), (1 + 1j))

    def test_parse_single(self):
        self.assertEqual(
            self.ureg.parse_expression("meter"), self.Q_(1, UnitsContainer(meter=1.0))
        )
        self.assertEqual(
            self.ureg.parse_expression("second"), self.Q_(1, UnitsContainer(second=1.0))
        )

    def test_parse_alias(self):
        self.assertEqual(
            self.ureg.parse_expression("metre"), self.Q_(1, UnitsContainer(meter=1.0))
        )

    def test_parse_plural(self):
        self.assertEqual(
            self.ureg.parse_expression("meters"), self.Q_(1, UnitsContainer(meter=1.0))
        )

    def test_parse_prefix(self):
        self.assertEqual(
            self.ureg.parse_expression("kilometer"),
            self.Q_(1, UnitsContainer(kilometer=1.0)),
        )

    def test_parse_complex(self):
        self.assertEqual(
            self.ureg.parse_expression("kilometre"),
            self.Q_(1, UnitsContainer(kilometer=1.0)),
        )
        self.assertEqual(
            self.ureg.parse_expression("kilometres"),
            self.Q_(1, UnitsContainer(kilometer=1.0)),
        )

    def test_parse_mul_div(self):
        self.assertEqual(
            self.ureg.parse_expression("meter*meter"),
            self.Q_(1, UnitsContainer(meter=2.0)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter**2"),
            self.Q_(1, UnitsContainer(meter=2.0)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter*second"),
            self.Q_(1, UnitsContainer(meter=1.0, second=1)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter/second"),
            self.Q_(1, UnitsContainer(meter=1.0, second=-1)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter/second**2"),
            self.Q_(1, UnitsContainer(meter=1.0, second=-2)),
        )

    def test_parse_pretty(self):
        self.assertEqual(
            self.ureg.parse_expression("meter/second²"),
            self.Q_(1, UnitsContainer(meter=1.0, second=-2)),
        )
        self.assertEqual(
            self.ureg.parse_expression("m³/s³"),
            self.Q_(1, UnitsContainer(meter=3.0, second=-3)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter² · second"),
            self.Q_(1, UnitsContainer(meter=2.0, second=1)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter⁰.⁵·second"),
            self.Q_(1, UnitsContainer(meter=0.5, second=1)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter³⁷/second⁴.³²¹"),
            self.Q_(1, UnitsContainer(meter=37, second=-4.321)),
        )

    def test_parse_factor(self):
        self.assertEqual(
            self.ureg.parse_expression("42*meter"),
            self.Q_(42, UnitsContainer(meter=1.0)),
        )
        self.assertEqual(
            self.ureg.parse_expression("meter*42"),
            self.Q_(42, UnitsContainer(meter=1.0)),
        )

    def test_rep_and_parse(self):
        q = self.Q_(1, "g/(m**2*s)")
        self.assertEqual(self.Q_(q.magnitude, str(q.units)), q)

    def test_as_delta(self):
        parse = self.ureg.parse_units
        self.assertEqual(parse("kelvin", as_delta=True), UnitsContainer(kelvin=1))
        self.assertEqual(parse("kelvin", as_delta=False), UnitsContainer(kelvin=1))
        self.assertEqual(
            parse("kelvin**(-1)", as_delta=True), UnitsContainer(kelvin=-1)
        )
        self.assertEqual(
            parse("kelvin**(-1)", as_delta=False), UnitsContainer(kelvin=-1)
        )
        self.assertEqual(parse("kelvin**2", as_delta=True), UnitsContainer(kelvin=2))
        self.assertEqual(parse("kelvin**2", as_delta=False), UnitsContainer(kelvin=2))
        self.assertEqual(
            parse("kelvin*meter", as_delta=True), UnitsContainer(kelvin=1, meter=1)
        )
        self.assertEqual(
            parse("kelvin*meter", as_delta=False), UnitsContainer(kelvin=1, meter=1)
        )

    def test_parse_expression_with_preprocessor(self):
        # Add parsing of UDUNITS-style power
        self.ureg.preprocessors.append(
            functools.partial(
                re.sub,
                r"(?<=[A-Za-z])(?![A-Za-z])(?<![0-9\-][eE])(?<![0-9\-])(?=[0-9\-])",
                "**",
            )
        )
        # Test equality
        self.assertEqual(
            self.ureg.parse_expression("42 m2"), self.Q_(42, UnitsContainer(meter=2.0))
        )
        self.assertEqual(
            self.ureg.parse_expression("1e6 Hz s-2"),
            self.Q_(1e6, UnitsContainer(second=-3.0)),
        )
        self.assertEqual(
            self.ureg.parse_expression("3 metre3"),
            self.Q_(3, UnitsContainer(meter=3.0)),
        )
        # Clean up and test previously expected value
        self.ureg.preprocessors.pop()
        self.assertEqual(
            self.ureg.parse_expression("1e6 Hz s-2"),
            self.Q_(999998.0, UnitsContainer()),
        )

    def test_parse_unit_with_preprocessor(self):
        # Add parsing of UDUNITS-style power
        self.ureg.preprocessors.append(
            functools.partial(
                re.sub,
                r"(?<=[A-Za-z])(?![A-Za-z])(?<![0-9\-][eE])(?<![0-9\-])(?=[0-9\-])",
                "**",
            )
        )
        # Test equality
        self.assertEqual(self.ureg.parse_units("m2"), UnitsContainer(meter=2.0))
        self.assertEqual(self.ureg.parse_units("m-2"), UnitsContainer(meter=-2.0))
        # Clean up
        self.ureg.preprocessors.pop()

    def test_name(self):
        self.assertRaises(UndefinedUnitError, self.ureg.get_name, "asdf")

    def test_symbol(self):
        self.assertRaises(UndefinedUnitError, self.ureg.get_symbol, "asdf")

        self.assertEqual(self.ureg.get_symbol("meter"), "m")
        self.assertEqual(self.ureg.get_symbol("second"), "s")
        self.assertEqual(self.ureg.get_symbol("hertz"), "Hz")

        self.assertEqual(self.ureg.get_symbol("kilometer"), "km")
        self.assertEqual(self.ureg.get_symbol("megahertz"), "MHz")
        self.assertEqual(self.ureg.get_symbol("millisecond"), "ms")

    def test_imperial_symbol(self):
        self.assertEqual(self.ureg.get_symbol("inch"), "in")
        self.assertEqual(self.ureg.get_symbol("foot"), "ft")
        self.assertEqual(self.ureg.get_symbol("inches"), "in")
        self.assertEqual(self.ureg.get_symbol("feet"), "ft")
        self.assertEqual(self.ureg.get_symbol("international_foot"), "ft")
        self.assertEqual(self.ureg.get_symbol("international_inch"), "in")

    def test_pint(self):
        self.assertLess(self.ureg.pint, self.ureg.liter)
        self.assertLess(self.ureg.pint, self.ureg.imperial_pint)

    def test_wraps(self):
        def func(x):
            return x

        ureg = self.ureg

        self.assertRaises(TypeError, ureg.wraps, (3 * ureg.meter, [None]))
        self.assertRaises(TypeError, ureg.wraps, (None, [3 * ureg.meter]))

        f0 = ureg.wraps(None, [None])(func)
        self.assertEqual(f0(3.0), 3.0)

        f0 = ureg.wraps(None, None)(func)
        self.assertEqual(f0(3.0), 3.0)

        f1 = ureg.wraps(None, ["meter"])(func)
        self.assertRaises(ValueError, f1, 3.0)
        self.assertEqual(f1(3.0 * ureg.centimeter), 0.03)
        self.assertEqual(f1(3.0 * ureg.meter), 3.0)
        self.assertRaises(DimensionalityError, f1, 3 * ureg.second)

        f1b = ureg.wraps(None, [ureg.meter])(func)
        self.assertRaises(ValueError, f1b, 3.0)
        self.assertEqual(f1b(3.0 * ureg.centimeter), 0.03)
        self.assertEqual(f1b(3.0 * ureg.meter), 3.0)
        self.assertRaises(DimensionalityError, f1b, 3 * ureg.second)

        f1c = ureg.wraps("meter", [ureg.meter])(func)
        self.assertEqual(f1c(3.0 * ureg.centimeter), 0.03 * ureg.meter)
        self.assertEqual(f1c(3.0 * ureg.meter), 3.0 * ureg.meter)
        self.assertRaises(DimensionalityError, f1c, 3 * ureg.second)

        f1d = ureg.wraps(ureg.meter, [ureg.meter])(func)
        self.assertEqual(f1d(3.0 * ureg.centimeter), 0.03 * ureg.meter)
        self.assertEqual(f1d(3.0 * ureg.meter), 3.0 * ureg.meter)
        self.assertRaises(DimensionalityError, f1d, 3 * ureg.second)

        f1 = ureg.wraps(None, "meter")(func)
        self.assertRaises(ValueError, f1, 3.0)
        self.assertEqual(f1(3.0 * ureg.centimeter), 0.03)
        self.assertEqual(f1(3.0 * ureg.meter), 3.0)
        self.assertRaises(DimensionalityError, f1, 3 * ureg.second)

        f2 = ureg.wraps("centimeter", ["meter"])(func)
        self.assertRaises(ValueError, f2, 3.0)
        self.assertEqual(f2(3.0 * ureg.centimeter), 0.03 * ureg.centimeter)
        self.assertEqual(f2(3.0 * ureg.meter), 3 * ureg.centimeter)

        f3 = ureg.wraps("centimeter", ["meter"], strict=False)(func)
        self.assertEqual(f3(3), 3 * ureg.centimeter)
        self.assertEqual(f3(3.0 * ureg.centimeter), 0.03 * ureg.centimeter)
        self.assertEqual(f3(3.0 * ureg.meter), 3.0 * ureg.centimeter)

        def gfunc(x, y):
            return x + y

        g0 = ureg.wraps(None, [None, None])(gfunc)
        self.assertEqual(g0(3, 1), 4)

        g1 = ureg.wraps(None, ["meter", "centimeter"])(gfunc)
        self.assertRaises(ValueError, g1, 3 * ureg.meter, 1)
        self.assertEqual(g1(3 * ureg.meter, 1 * ureg.centimeter), 4)
        self.assertEqual(g1(3 * ureg.meter, 1 * ureg.meter), 3 + 100)

        def hfunc(x, y):
            return x, y

        h0 = ureg.wraps(None, [None, None])(hfunc)
        self.assertEqual(h0(3, 1), (3, 1))

        h1 = ureg.wraps(["meter", "centimeter"], [None, None])(hfunc)
        self.assertEqual(h1(3, 1), [3 * ureg.meter, 1 * ureg.cm])

        h2 = ureg.wraps(("meter", "centimeter"), [None, None])(hfunc)
        self.assertEqual(h2(3, 1), (3 * ureg.meter, 1 * ureg.cm))

        h3 = ureg.wraps((None,), (None, None))(hfunc)
        self.assertEqual(h3(3, 1), (3, 1))

    def test_wrap_referencing(self):

        ureg = self.ureg

        def gfunc(x, y):
            return x + y

        def gfunc2(x, y):
            return x ** 2 + y

        def gfunc3(x, y):
            return x ** 2 * y

        rst = 3.0 * ureg.meter + 1.0 * ureg.centimeter

        g0 = ureg.wraps("=A", ["=A", "=A"])(gfunc)
        self.assertEqual(g0(3.0 * ureg.meter, 1.0 * ureg.centimeter), rst.to("meter"))
        self.assertEqual(g0(3, 1), 4)

        g1 = ureg.wraps("=A", ["=A", "=A"])(gfunc)
        self.assertEqual(
            g1(3.0 * ureg.meter, 1.0 * ureg.centimeter), rst.to("centimeter")
        )

        g2 = ureg.wraps("=A", ["=A", "=A"])(gfunc)
        self.assertEqual(g2(3.0 * ureg.meter, 1.0 * ureg.centimeter), rst.to("meter"))

        g3 = ureg.wraps("=A**2", ["=A", "=A**2"])(gfunc2)
        a = 3.0 * ureg.meter
        b = (2.0 * ureg.centimeter) ** 2
        self.assertEqual(g3(a, b), gfunc2(a, b))
        self.assertEqual(g3(3, 2), gfunc2(3, 2))

        g4 = ureg.wraps("=A**2 * B", ["=A", "=B"])(gfunc3)
        self.assertEqual(
            g4(3.0 * ureg.meter, 2.0 * ureg.second), ureg("(3*meter)**2 * 2 *second")
        )
        self.assertEqual(g4(3.0 * ureg.meter, 2.0), ureg("(3*meter)**2 * 2"))
        self.assertEqual(g4(3.0, 2.0 * ureg.second), ureg("3**2 * 2 * second"))

    def test_check(self):
        def func(x):
            return x

        ureg = self.ureg

        f0 = ureg.check("[length]")(func)
        self.assertRaises(DimensionalityError, f0, 3.0)
        self.assertEqual(f0(3.0 * ureg.centimeter), 0.03 * ureg.meter)
        self.assertRaises(DimensionalityError, f0, 3.0 * ureg.kilogram)

        f0b = ureg.check(ureg.meter)(func)
        self.assertRaises(DimensionalityError, f0b, 3.0)
        self.assertEqual(f0b(3.0 * ureg.centimeter), 0.03 * ureg.meter)
        self.assertRaises(DimensionalityError, f0b, 3.0 * ureg.kilogram)

        def gfunc(x, y):
            return x / y

        g0 = ureg.check(None, None)(gfunc)
        self.assertEqual(g0(6, 2), 3)
        self.assertEqual(g0(6 * ureg.parsec, 2), 3 * ureg.parsec)

        g1 = ureg.check("[speed]", "[time]")(gfunc)
        self.assertRaises(DimensionalityError, g1, 3.0, 1)
        self.assertRaises(DimensionalityError, g1, 1 * ureg.parsec, 1 * ureg.angstrom)
        self.assertRaises(TypeError, g1, 1 * ureg.km / ureg.hour, 1 * ureg.hour, 3.0)
        self.assertEqual(
            g1(3.6 * ureg.km / ureg.hour, 1 * ureg.second),
            1 * ureg.meter / ureg.second ** 2,
        )

        self.assertRaises(TypeError, ureg.check("[speed]"), gfunc)
        self.assertRaises(TypeError, ureg.check("[speed]", "[time]", "[mass]"), gfunc)

    def test_to_ref_vs_to(self):
        self.ureg.autoconvert_offset_to_baseunit = True
        q = 8.0 * self.ureg.inch
        t = 8.0 * self.ureg.degF
        dt = 8.0 * self.ureg.delta_degF
        self.assertEqual(
            q.to("yard").magnitude, self.ureg._units["inch"].converter.to_reference(8.0)
        )
        self.assertEqual(
            t.to("kelvin").magnitude,
            self.ureg._units["degF"].converter.to_reference(8.0),
        )
        self.assertEqual(
            dt.to("kelvin").magnitude,
            self.ureg._units["delta_degF"].converter.to_reference(8.0),
        )

    def test_redefinition(self):
        d = UnitRegistry().define

        with self.capture_log() as buffer:
            d("meter = [fruits]")
            d("kilo- = 1000")
            d("[speed] = [vegetables]")

            # aliases
            d("bla = 3.2 meter = inch")
            d("myk- = 1000 = kilo-")

            self.assertEqual(len(buffer), 5)

    def test_convert_parse_str(self):
        ureg = self.ureg
        self.assertEqual(
            ureg.convert(1, "meter", "inch"),
            ureg.convert(1, UnitsContainer(meter=1), UnitsContainer(inch=1)),
        )

    @helpers.requires_numpy()
    def test_convert_inplace(self):
        ureg = self.ureg

        # Conversions with single units take a different codepath than
        # Conversions with more than one unit.
        src_dst1 = UnitsContainer(meter=1), UnitsContainer(inch=1)
        src_dst2 = UnitsContainer(meter=1, second=-1), UnitsContainer(inch=1, minute=-1)
        for src, dst in (src_dst1, src_dst2):
            v = (ureg.convert(1, src, dst),)

            a = np.ones((3, 1))
            ac = np.ones((3, 1))

            r1 = ureg.convert(a, src, dst)
            np.testing.assert_allclose(r1, v * ac)
            self.assertIsNot(r1, a)

            r2 = ureg.convert(a, src, dst, inplace=True)
            np.testing.assert_allclose(r2, v * ac)
            self.assertIs(r2, a)

    def test_repeated_convert(self):
        # Because of caching, repeated conversions were failing.
        self.ureg.convert(1, "m", "ft")
        self.ureg.convert(1, "m", "ft")

    def test_singular_SI_prefix_convert(self):
        # Fix for issue 156
        self.ureg.convert(1, "mm", "m")
        self.ureg.convert(1, "ms", "s")
        self.ureg.convert(1, "m", "mm")
        self.ureg.convert(1, "s", "ms")

    def test_parse_units(self):
        ureg = self.ureg
        self.assertEqual(ureg.parse_units(""), ureg.Unit(""))
        self.assertRaises(ValueError, ureg.parse_units, "2 * meter")


class TestCompatibleUnits(QuantityTestCase):
    FORCE_NDARRAY = False

    def setUp(self):
        super().setUp()
        self.ureg = UnitRegistry(force_ndarray=self.FORCE_NDARRAY)
        self.Q_ = self.ureg.Quantity
        self.U_ = self.ureg.Unit

    def _test(self, input_units):
        gd = self.ureg.get_dimensionality
        dim = gd(input_units)
        equiv = self.ureg.get_compatible_units(input_units)
        for eq in equiv:
            self.assertEqual(gd(eq), dim)
        self.assertEqual(equiv, self.ureg.get_compatible_units(dim))

    def _test2(self, units1, units2):
        equiv1 = self.ureg.get_compatible_units(units1)
        equiv2 = self.ureg.get_compatible_units(units2)
        self.assertEqual(equiv1, equiv2)

    def test_many(self):
        self._test(self.ureg.meter)
        self._test(self.ureg.seconds)
        self._test(self.ureg.newton)
        self._test(self.ureg.kelvin)

    def test_context_sp(self):

        gd = self.ureg.get_dimensionality

        # length, frequency, energy
        valid = [
            gd(self.ureg.meter),
            gd(self.ureg.hertz),
            gd(self.ureg.joule),
            1 / gd(self.ureg.meter),
        ]

        with self.ureg.context("sp"):
            equiv = self.ureg.get_compatible_units(self.ureg.meter)
            result = set()
            for eq in equiv:
                dim = gd(eq)
                result.add(dim)
                self.assertIn(dim, valid)

            self.assertEqual(len(result), len(valid))

    def test_get_base_units(self):
        ureg = UnitRegistry()
        self.assertEqual(ureg.get_base_units(""), (1, ureg.Unit("")))
        self.assertEqual(ureg.get_base_units("pi"), (math.pi, ureg.Unit("")))
        self.assertEqual(ureg.get_base_units("ln10"), (math.log(10), ureg.Unit("")))
        self.assertEqual(
            ureg.get_base_units("meter"), ureg.get_base_units(ParserHelper(meter=1))
        )

    def test_get_compatible_units(self):
        ureg = UnitRegistry()
        self.assertEqual(ureg.get_compatible_units(""), frozenset())
        self.assertEqual(
            ureg.get_compatible_units("meter"),
            ureg.get_compatible_units(ParserHelper(meter=1)),
        )


class TestRegistryWithDefaultRegistry(TestRegistry):
    @classmethod
    def setUpClass(cls):
        from pint import _DEFAULT_REGISTRY

        cls.ureg = _DEFAULT_REGISTRY
        cls.Q_ = cls.ureg.Quantity

    def test_lazy(self):
        x = LazyRegistry()
        x.test = "test"
        self.assertIsInstance(x, UnitRegistry)
        y = LazyRegistry()
        y("meter")
        self.assertIsInstance(y, UnitRegistry)

    def test_redefinition(self):
        d = self.ureg.define
        self.assertRaises(DefinitionSyntaxError, d, "meter = [time]")
        self.assertRaises(RedefinitionError, d, "meter = [newdim]")
        self.assertRaises(RedefinitionError, d, "kilo- = 1000")
        self.assertRaises(RedefinitionError, d, "[speed] = [length]")

        # aliases
        self.assertIn("inch", self.ureg._units)
        self.assertRaises(RedefinitionError, d, "bla = 3.2 meter = inch")
        self.assertRaises(RedefinitionError, d, "myk- = 1000 = kilo-")


class TestConvertWithOffset(QuantityTestCase, ParameterizedTestCase):

    # The dicts in convert_with_offset are used to create a UnitsContainer.
    # We create UnitsContainer to avoid any auto-conversion of units.
    convert_with_offset = [
        (({"degC": 1}, {"degC": 1}), 10),
        (({"degC": 1}, {"kelvin": 1}), 283.15),
        (({"degC": 1}, {"degC": 1, "millimeter": 1, "meter": -1}), "error"),
        (({"degC": 1}, {"kelvin": 1, "millimeter": 1, "meter": -1}), 283150),
        (({"kelvin": 1}, {"degC": 1}), -263.15),
        (({"kelvin": 1}, {"kelvin": 1}), 10),
        (({"kelvin": 1}, {"degC": 1, "millimeter": 1, "meter": -1}), "error"),
        (({"kelvin": 1}, {"kelvin": 1, "millimeter": 1, "meter": -1}), 10000),
        (({"degC": 1, "millimeter": 1, "meter": -1}, {"degC": 1}), "error"),
        (({"degC": 1, "millimeter": 1, "meter": -1}, {"kelvin": 1}), "error"),
        (
            (
                {"degC": 1, "millimeter": 1, "meter": -1},
                {"degC": 1, "millimeter": 1, "meter": -1},
            ),
            10,
        ),
        (
            (
                {"degC": 1, "millimeter": 1, "meter": -1},
                {"kelvin": 1, "millimeter": 1, "meter": -1},
            ),
            "error",
        ),
        (({"kelvin": 1, "millimeter": 1, "meter": -1}, {"degC": 1}), -273.14),
        (({"kelvin": 1, "millimeter": 1, "meter": -1}, {"kelvin": 1}), 0.01),
        (
            (
                {"kelvin": 1, "millimeter": 1, "meter": -1},
                {"degC": 1, "millimeter": 1, "meter": -1},
            ),
            "error",
        ),
        (
            (
                {"kelvin": 1, "millimeter": 1, "meter": -1},
                {"kelvin": 1, "millimeter": 1, "meter": -1},
            ),
            10,
        ),
        (({"degC": 2}, {"kelvin": 2}), "error"),
        (({"degC": 1, "degF": 1}, {"kelvin": 2}), "error"),
        (({"degC": 1, "kelvin": 1}, {"kelvin": 2}), "error"),
    ]

    @ParameterizedTestCase.parameterize(
        ("input", "expected_output"), convert_with_offset
    )
    def test_to_and_from_offset_units(self, input_tuple, expected):
        src, dst = input_tuple
        src, dst = UnitsContainer(src), UnitsContainer(dst)
        value = 10.0
        convert = self.ureg.convert
        if isinstance(expected, str):
            self.assertRaises(DimensionalityError, convert, value, src, dst)
            if src != dst:
                self.assertRaises(DimensionalityError, convert, value, dst, src)
        else:
            self.assertQuantityAlmostEqual(
                convert(value, src, dst), expected, atol=0.001
            )
            if src != dst:
                self.assertQuantityAlmostEqual(
                    convert(expected, dst, src), value, atol=0.001
                )

    def test_alias(self):
        # Use load_definitions
        ureg = UnitRegistry(
            [
                "canonical = [] = can = alias1 = alias2\n",
                # overlapping aliases
                "@alias canonical = alias2 = alias3\n",
                # Against another alias
                "@alias alias3 = alias4\n",
            ]
        )

        # Use define
        ureg.define("@alias canonical = alias5")

        # Test that new aliases work
        # Test that pre-existing aliases and symbol are not eliminated
        for a in ("can", "alias1", "alias2", "alias3", "alias4", "alias5"):
            self.assertEqual(ureg.Unit(a), ureg.Unit("canonical"))

        # Test that aliases defined multiple times are not duplicated
        self.assertEqual(
            ureg._units["canonical"].aliases,
            ("alias1", "alias2", "alias3", "alias4", "alias5"),
        )

        # Define against unknown name
        self.assertRaises(KeyError, ureg.define, "@alias notexist = something")



import pickle

from pint import (
    DefinitionSyntaxError,
    DimensionalityError,
    OffsetUnitCalculusError,
    Quantity,
    RedefinitionError,
    UndefinedUnitError,
    UnitRegistry,
)
from pint.testsuite import BaseTestCase


class TestErrors(BaseTestCase):
    def test_definition_syntax_error(self):
        ex = DefinitionSyntaxError("foo")
        self.assertEqual(str(ex), "foo")

        # filename and lineno can be attached after init
        ex.filename = "a.txt"
        ex.lineno = 123
        self.assertEqual(str(ex), "While opening a.txt, in line 123: foo")

        ex = DefinitionSyntaxError("foo", lineno=123)
        self.assertEqual(str(ex), "In line 123: foo")

        ex = DefinitionSyntaxError("foo", filename="a.txt")
        self.assertEqual(str(ex), "While opening a.txt: foo")

        ex = DefinitionSyntaxError("foo", filename="a.txt", lineno=123)
        self.assertEqual(str(ex), "While opening a.txt, in line 123: foo")

    def test_redefinition_error(self):
        ex = RedefinitionError("foo", "bar")
        self.assertEqual(str(ex), "Cannot redefine 'foo' (bar)")

        # filename and lineno can be attached after init
        ex.filename = "a.txt"
        ex.lineno = 123
        self.assertEqual(
            str(ex), "While opening a.txt, in line 123: Cannot redefine 'foo' (bar)"
        )

        ex = RedefinitionError("foo", "bar", lineno=123)
        self.assertEqual(str(ex), "In line 123: Cannot redefine 'foo' (bar)")

        ex = RedefinitionError("foo", "bar", filename="a.txt")
        self.assertEqual(str(ex), "While opening a.txt: Cannot redefine 'foo' (bar)")

        ex = RedefinitionError("foo", "bar", filename="a.txt", lineno=123)
        self.assertEqual(
            str(ex), "While opening a.txt, in line 123: Cannot redefine 'foo' (bar)"
        )

    def test_undefined_unit_error(self):
        x = ("meter",)
        msg = "'meter' is not defined in the unit registry"
        self.assertEqual(str(UndefinedUnitError(x)), msg)
        self.assertEqual(str(UndefinedUnitError(list(x))), msg)
        self.assertEqual(str(UndefinedUnitError(set(x))), msg)

    def test_undefined_unit_error_multi(self):
        x = ("meter", "kg")
        msg = "('meter', 'kg') are not defined in the unit registry"
        self.assertEqual(str(UndefinedUnitError(x)), msg)
        self.assertEqual(str(UndefinedUnitError(list(x))), msg)

    def test_dimensionality_error(self):
        ex = DimensionalityError("a", "b")
        self.assertEqual(str(ex), "Cannot convert from 'a' to 'b'")
        ex = DimensionalityError("a", "b", "c")
        self.assertEqual(str(ex), "Cannot convert from 'a' (c) to 'b' ()")
        ex = DimensionalityError("a", "b", "c", "d", extra_msg=": msg")
        self.assertEqual(str(ex), "Cannot convert from 'a' (c) to 'b' (d): msg")

    def test_offset_unit_calculus_error(self):
        ex = OffsetUnitCalculusError(Quantity("1 kg")._units)
        self.assertEqual(
            str(ex),
            "Ambiguous operation with offset unit (kilogram). See "
            "https://pint.readthedocs.io/en/latest/nonmult.html for guidance.",
        )
        ex = OffsetUnitCalculusError(Quantity("1 kg")._units, Quantity("1 s")._units)
        self.assertEqual(
            str(ex),
            "Ambiguous operation with offset unit (kilogram, second). See "
            "https://pint.readthedocs.io/en/latest/nonmult.html for guidance.",
        )

    def test_pickle_definition_syntax_error(self):
        # OffsetUnitCalculusError raised from a custom ureg must be pickleable even if
        # the ureg is not registered as the application ureg
        ureg = UnitRegistry(filename=None)
        ureg.define("foo = [bar]")
        ureg.define("bar = 2 foo")
        pik = pickle.dumps(ureg.Quantity("1 foo"))
        with self.assertRaises(UndefinedUnitError):
            pickle.loads(pik)
        q1 = ureg.Quantity("1 foo")
        q2 = ureg.Quantity("1 bar")

        for ex in [
            DefinitionSyntaxError("foo", filename="a.txt", lineno=123),
            RedefinitionError("foo", "bar"),
            UndefinedUnitError("meter"),
            DimensionalityError("a", "b", "c", "d", extra_msg=": msg"),
            OffsetUnitCalculusError(Quantity("1 kg")._units, Quantity("1 s")._units),
            OffsetUnitCalculusError(q1._units, q2._units),
        ]:
            with self.subTest(etype=type(ex)):
                # assert False, ex.__reduce__()
                ex2 = pickle.loads(pickle.dumps(ex))
                assert type(ex) is type(ex2)
                self.assertEqual(ex.args, ex2.args)
                self.assertEqual(ex.__dict__, ex2.__dict__)
                self.assertEqual(str(ex), str(ex2))



"""
    pint.context
    ~~~~~~~~~~~~

    Functions and classes related to context definitions and application.

    :copyright: 2016 by Pint Authors, see AUTHORS for more details..
    :license: BSD, see LICENSE for more details.
"""

import re
import weakref
from collections import ChainMap, defaultdict

from .definitions import Definition, UnitDefinition
from .errors import DefinitionSyntaxError
from .util import ParserHelper, SourceIterator, to_units_container

#: Regex to match the header parts of a context.
_header_re = re.compile(
    r"@context\s*(?P<defaults>\(.*\))?\s+(?P<name>\w+)\s*(=(?P<aliases>.*))*"
)

#: Regex to match variable names in an equation.
_varname_re = re.compile(r"[A-Za-z_][A-Za-z0-9_]*")


def _expression_to_function(eq):
    def func(ureg, value, **kwargs):
        return ureg.parse_expression(eq, value=value, **kwargs)

    return func


class Context:
    """A specialized container that defines transformation functions from one
    dimension to another. Each Dimension are specified using a UnitsContainer.
    Simple transformation are given with a function taking a single parameter.


    Conversion functions may take optional keyword arguments and the context
    can have default values for these arguments.


    Additionally, a context may host redefinitions:


    A redefinition must be performed among units that already exist in the registry. It
    cannot change the dimensionality of a unit. The symbol and aliases are automatically
    inherited from the registry.

    Parameters
    ----------
    name : str or None, optional
        Name of the context (must be unique within the registry).
        Use None for anonymous Context. (Default value = None).
    aliases : iterable of str
        Other names for the context.
    defaults : None or dict
        Maps variable names to values.

    Example
    -------

    >>> from pint.util import UnitsContainer
    >>> timedim = UnitsContainer({'[time]': 1})
    >>> spacedim = UnitsContainer({'[length]': 1})
    >>> def f(time):
    ...     'Time to length converter'
    ...     return 3. * time
    >>> c = Context()
    >>> c.add_transformation(timedim, spacedim, f)
    >>> c.transform(timedim, spacedim, 2)
    6
    >>> def f(time, n):
    ...     'Time to length converter, n is the index of refraction of the material'
    ...     return 3. * time / n
    >>> c = Context(n=3)
    >>> c.add_transformation(timedim, spacedim, f)
    >>> c.transform(timedim, spacedim, 2)
    2
    >>> c.redefine("pound = 0.5 kg")
    """

    def __init__(self, name=None, aliases=(), defaults=None):

        self.name = name
        self.aliases = aliases

        #: Maps (src, dst) -> transformation function
        self.funcs = {}

        #: Maps defaults variable names to values
        self.defaults = defaults or {}

        # Store Definition objects that are context-specific
        self.redefinitions = []

        # Flag set to True by the Registry the first time the context is enabled
        self.checked = False

        #: Maps (src, dst) -> self
        #: Used as a convenience dictionary to be composed by ContextChain
        self.relation_to_context = weakref.WeakValueDictionary()

    @classmethod
    def from_context(cls, context, **defaults):
        """Creates a new context that shares the funcs dictionary with the
        original context. The default values are copied from the original
        context and updated with the new defaults.

        If defaults is empty, return the same context.

        Parameters
        ----------
        context : pint.Context
            Original context.
        **defaults


        Returns
        -------
        pint.Context
        """
        if defaults:
            newdef = dict(context.defaults, **defaults)
            c = cls(context.name, context.aliases, newdef)
            c.funcs = context.funcs
            c.redefinitions = context.redefinitions
            for edge in context.funcs:
                c.relation_to_context[edge] = c
            return c
        return context

    @classmethod
    def from_lines(cls, lines, to_base_func=None):
        lines = SourceIterator(lines)

        lineno, header = next(lines)
        try:
            r = _header_re.search(header)
            name = r.groupdict()["name"].strip()
            aliases = r.groupdict()["aliases"]
            if aliases:
                aliases = tuple(a.strip() for a in r.groupdict()["aliases"].split("="))
            else:
                aliases = ()
            defaults = r.groupdict()["defaults"]
        except Exception as exc:
            raise DefinitionSyntaxError(
                "Could not parse the Context header '%s'" % header, lineno=lineno
            ) from exc

        if defaults:

            def to_num(val):
                val = complex(val)
                if not val.imag:
                    return val.real
                return val

            txt = defaults
            try:
                defaults = (part.split("=") for part in defaults.strip("()").split(","))
                defaults = {str(k).strip(): to_num(v) for k, v in defaults}
            except (ValueError, TypeError) as exc:
                raise DefinitionSyntaxError(
                    f"Could not parse Context definition defaults: '{txt}'",
                    lineno=lineno,
                ) from exc

            ctx = cls(name, aliases, defaults)
        else:
            ctx = cls(name, aliases)

        names = set()
        for lineno, line in lines:
            if "=" in line:
                ctx.redefine(line)
                continue

            try:
                rel, eq = line.split(":")
                names.update(_varname_re.findall(eq))

                func = _expression_to_function(eq)

                if "<->" in rel:
                    src, dst = (ParserHelper.from_string(s) for s in rel.split("<->"))
                    if to_base_func:
                        src = to_base_func(src)
                        dst = to_base_func(dst)
                    ctx.add_transformation(src, dst, func)
                    ctx.add_transformation(dst, src, func)
                elif "->" in rel:
                    src, dst = (ParserHelper.from_string(s) for s in rel.split("->"))
                    if to_base_func:
                        src = to_base_func(src)
                        dst = to_base_func(dst)
                    ctx.add_transformation(src, dst, func)
                else:
                    raise Exception
            except Exception as exc:
                raise DefinitionSyntaxError(
                    "Could not parse Context %s relation '%s'" % (name, line),
                    lineno=lineno,
                ) from exc

        if defaults:
            missing_pars = defaults.keys() - set(names)
            if missing_pars:
                raise DefinitionSyntaxError(
                    f"Context parameters {missing_pars} not found in any equation"
                )

        return ctx

    def add_transformation(self, src, dst, func):
        """Add a transformation function to the context.
        """

        _key = self.__keytransform__(src, dst)
        self.funcs[_key] = func
        self.relation_to_context[_key] = self

    def remove_transformation(self, src, dst):
        """Add a transformation function to the context.
        """

        _key = self.__keytransform__(src, dst)
        del self.funcs[_key]
        del self.relation_to_context[_key]

    @staticmethod
    def __keytransform__(src, dst):
        return to_units_container(src), to_units_container(dst)

    def transform(self, src, dst, registry, value):
        """Transform a value.
        """

        _key = self.__keytransform__(src, dst)
        return self.funcs[_key](registry, value, **self.defaults)

    def redefine(self, definition: str) -> None:
        """Override the definition of a unit in the registry.

        Parameters
        ----------
        definition : str
            <unit> = <new definition>``, e.g. ``pound = 0.5 kg``
        """

        for line in definition.splitlines():
            d = Definition.from_string(line)
            if not isinstance(d, UnitDefinition):
                raise DefinitionSyntaxError(
                    "Expected <unit> = <converter>; got %s" % line.strip()
                )
            if d.symbol != d.name or d.aliases:
                raise DefinitionSyntaxError(
                    "Can't change a unit's symbol or aliases within a context"
                )
            if d.is_base:
                raise DefinitionSyntaxError("Can't define base units within a context")
            self.redefinitions.append(d)

    def hashable(self):
        """Generate a unique hashable and comparable representation of self, which can
        be used as a key in a dict. This class cannot define ``__hash__`` because it is
        mutable, and the Python interpreter does cache the output of ``__hash__``.

        Returns
        -------
        tuple
        """
        return (
            self.name,
            tuple(self.aliases),
            frozenset((k, id(v)) for k, v in self.funcs.items()),
            frozenset(self.defaults.items()),
            tuple(self.redefinitions),
        )


class ContextChain(ChainMap):
    """A specialized ChainMap for contexts that simplifies finding rules
    to transform from one dimension to another.
    """

    def __init__(self):
        super().__init__()
        self.contexts = []
        self.maps.clear()  # Remove default empty map
        self._graph = None

    def insert_contexts(self, *contexts):
        """Insert one or more contexts in reversed order the chained map.
        (A rule in last context will take precedence)

        To facilitate the identification of the context with the matching rule,
        the *relation_to_context* dictionary of the context is used.
        """

        self.contexts = list(reversed(contexts)) + self.contexts
        self.maps = [ctx.relation_to_context for ctx in reversed(contexts)] + self.maps
        self._graph = None

    def remove_contexts(self, n: int = None):
        """Remove the last n inserted contexts from the chain.

        Parameters
        ----------
        n: int
            (Default value = None)
        """

        del self.contexts[:n]
        del self.maps[:n]
        self._graph = None

    @property
    def defaults(self):
        if self:
            return next(iter(self.maps[0].values())).defaults
        return {}

    @property
    def graph(self):
        """The graph relating"""
        if self._graph is None:
            self._graph = defaultdict(set)
            for fr_, to_ in self:
                self._graph[fr_].add(to_)
        return self._graph

    def transform(self, src, dst, registry, value):
        """Transform the value, finding the rule in the chained context.
        (A rule in last context will take precedence)
        """
        return self[(src, dst)].transform(src, dst, registry, value)

    def hashable(self):
        """Generate a unique hashable and comparable representation of self, which can
        be used as a key in a dict. This class cannot define ``__hash__`` because it is
        mutable, and the Python interpreter does cache the output of ``__hash__``.
        """
        return tuple(ctx.hashable() for ctx in self.contexts)



from pint import UnitRegistry
from pint.testsuite import QuantityTestCase


class TestGroup(QuantityTestCase):
    def _build_empty_reg_root(self):
        ureg = UnitRegistry(None)
        grp = ureg.get_group("root")
        grp.remove_units("pi")
        grp.invalidate_members()
        return ureg, ureg.get_group("root")

    def test_units_programatically(self):
        ureg, root = self._build_empty_reg_root()
        d = ureg._groups

        self.assertEqual(root._used_groups, set())
        self.assertEqual(root._used_by, set())
        root.add_units("meter", "second", "meter")
        self.assertEqual(root._unit_names, {"meter", "second"})
        self.assertEqual(root.members, {"meter", "second"})

        self.assertEqual(d.keys(), {"root"})

    def test_cyclic(self):
        ureg, root = self._build_empty_reg_root()
        g2 = ureg.Group("g2")
        g3 = ureg.Group("g3")
        g2.add_groups("g3")

        self.assertRaises(ValueError, g2.add_groups, "root")
        self.assertRaises(ValueError, g3.add_groups, "g2")
        self.assertRaises(ValueError, g3.add_groups, "root")

    def test_groups_programatically(self):
        ureg, root = self._build_empty_reg_root()
        d = ureg._groups
        g2 = ureg.Group("g2")

        self.assertEqual(d.keys(), {"root", "g2"})

        self.assertEqual(root._used_groups, {"g2"})
        self.assertEqual(root._used_by, set())

        self.assertEqual(g2._used_groups, set())
        self.assertEqual(g2._used_by, {"root"})

    def test_simple(self):
        lines = ["@group mygroup", "meter", "second"]

        ureg, root = self._build_empty_reg_root()
        d = ureg._groups

        grp = ureg.Group.from_lines(lines, lambda x: None)

        self.assertEqual(d.keys(), {"root", "mygroup"})

        self.assertEqual(grp.name, "mygroup")
        self.assertEqual(grp._unit_names, {"meter", "second"})
        self.assertEqual(grp._used_groups, set())
        self.assertEqual(grp._used_by, {root.name})
        self.assertEqual(grp.members, frozenset(["meter", "second"]))

    def test_using1(self):
        lines = ["@group mygroup using group1", "meter", "second"]

        ureg, root = self._build_empty_reg_root()
        ureg.Group("group1")
        grp = ureg.Group.from_lines(lines, lambda x: None)
        self.assertEqual(grp.name, "mygroup")
        self.assertEqual(grp._unit_names, {"meter", "second"})
        self.assertEqual(grp._used_groups, {"group1"})
        self.assertEqual(grp.members, frozenset(["meter", "second"]))

    def test_using2(self):
        lines = ["@group mygroup using group1,group2", "meter", "second"]

        ureg, root = self._build_empty_reg_root()
        ureg.Group("group1")
        ureg.Group("group2")
        grp = ureg.Group.from_lines(lines, lambda x: None)
        self.assertEqual(grp.name, "mygroup")
        self.assertEqual(grp._unit_names, {"meter", "second"})
        self.assertEqual(grp._used_groups, {"group1", "group2"})
        self.assertEqual(grp.members, frozenset(["meter", "second"]))

    def test_spaces(self):
        lines = ["@group   mygroup   using   group1 , group2", " meter ", " second  "]

        ureg, root = self._build_empty_reg_root()
        ureg.Group("group1")
        ureg.Group("group2")
        grp = ureg.Group.from_lines(lines, lambda x: None)
        self.assertEqual(grp.name, "mygroup")
        self.assertEqual(grp._unit_names, {"meter", "second"})
        self.assertEqual(grp._used_groups, {"group1", "group2"})
        self.assertEqual(grp.members, frozenset(["meter", "second"]))

    def test_invalidate_members(self):
        lines = ["@group mygroup using group1", "meter", "second"]

        ureg, root = self._build_empty_reg_root()
        ureg.Group("group1")
        grp = ureg.Group.from_lines(lines, lambda x: None)
        self.assertIs(root._computed_members, None)
        self.assertIs(grp._computed_members, None)
        self.assertEqual(grp.members, frozenset(["meter", "second"]))
        self.assertIs(root._computed_members, None)
        self.assertIsNot(grp._computed_members, None)
        self.assertEqual(root.members, frozenset(["meter", "second"]))
        self.assertIsNot(root._computed_members, None)
        self.assertIsNot(grp._computed_members, None)
        grp.invalidate_members()
        self.assertIs(root._computed_members, None)
        self.assertIs(grp._computed_members, None)

    def test_with_defintions(self):
        lines = [
            "@group imperial",
            "inch",
            "yard",
            "kings_leg = 2 * meter",
            "kings_head = 52 * inch" "pint",
        ]
        defs = []

        def define(ud):
            defs.append(ud.name)

        ureg, root = self._build_empty_reg_root()
        ureg.Group.from_lines(lines, define)

        self.assertEqual(["kings_leg", "kings_head"], defs)

    def test_members_including(self):
        ureg, root = self._build_empty_reg_root()

        g1 = ureg.Group("group1")
        g1.add_units("second", "inch")

        g2 = ureg.Group("group2")
        g2.add_units("second", "newton")

        g3 = ureg.Group("group3")
        g3.add_units("meter", "second")
        g3.add_groups("group1", "group2")

        self.assertEqual(root.members, frozenset(["meter", "second", "newton", "inch"]))
        self.assertEqual(g1.members, frozenset(["second", "inch"]))
        self.assertEqual(g2.members, frozenset(["second", "newton"]))
        self.assertEqual(g3.members, frozenset(["meter", "second", "newton", "inch"]))

    def test_get_compatible_units(self):
        ureg = UnitRegistry()

        g = ureg.get_group("test-imperial")
        g.add_units("inch", "yard", "pint")
        c = ureg.get_compatible_units("meter", "test-imperial")
        self.assertEqual(c, frozenset([ureg.inch, ureg.yard]))


class TestSystem(QuantityTestCase):
    def _build_empty_reg_root(self):
        ureg = UnitRegistry(None)
        grp = ureg.get_group("root")
        grp.remove_units("pi")
        grp.invalidate_members()
        return ureg, ureg.get_group("root")

    def test_implicit_root(self):
        lines = ["@system mks", "meter", "kilogram", "second"]

        ureg, root = self._build_empty_reg_root()
        s = ureg.System.from_lines(lines, lambda x: x)
        s._used_groups = {"root"}

    def test_simple_using(self):
        lines = ["@system mks using g1", "meter", "kilogram", "second"]

        ureg, root = self._build_empty_reg_root()
        s = ureg.System.from_lines(lines, lambda x: x)
        s._used_groups = {"root", "g1"}

    def test_members_group(self):
        lines = ["@system mk", "meter", "kilogram"]

        ureg, root = self._build_empty_reg_root()
        root.add_units("second")
        s = ureg.System.from_lines(lines, lambda x: x)
        self.assertEqual(s.members, frozenset(["second"]))

    def test_get_compatible_units(self):
        sysname = "mysys1"
        ureg = UnitRegistry()

        g = ureg.get_group("test-imperial")

        g.add_units("inch", "yard", "pint")
        c = ureg.get_compatible_units("meter", "test-imperial")
        self.assertEqual(c, frozenset([ureg.inch, ureg.yard]))

        lines = ["@system %s using test-imperial" % sysname, "inch"]

        ureg.System.from_lines(lines, lambda x: x)
        c = ureg.get_compatible_units("meter", sysname)
        self.assertEqual(c, frozenset([ureg.inch, ureg.yard]))

    def test_get_base_units(self):
        sysname = "mysys2"
        ureg = UnitRegistry()

        g = ureg.get_group("test-imperial")
        g.add_units("inch", "yard", "pint")

        lines = ["@system %s using test-imperial" % sysname, "inch"]

        s = ureg.System.from_lines(lines, ureg.get_base_units)
        ureg._systems[s.name] = s

        # base_factor, destination_units
        c = ureg.get_base_units("inch", system=sysname)
        self.assertAlmostEqual(c[0], 1)
        self.assertEqual(c[1], {"inch": 1})

        c = ureg.get_base_units("cm", system=sysname)
        self.assertAlmostEqual(c[0], 1.0 / 2.54)
        self.assertEqual(c[1], {"inch": 1})

    def test_get_base_units_different_exponent(self):
        sysname = "mysys3"
        ureg = UnitRegistry()

        g = ureg.get_group("test-imperial")
        g.add_units("inch", "yard", "pint")
        ureg.get_compatible_units("meter", "test-imperial")

        lines = ["@system %s using test-imperial" % sysname, "pint:meter"]

        s = ureg.System.from_lines(lines, ureg.get_base_units)
        ureg._systems[s.name] = s

        # base_factor, destination_units
        c = ureg.get_base_units("inch", system=sysname)
        self.assertAlmostEqual(c[0], 0.326, places=3)
        self.assertEqual(c[1], {"pint": 1.0 / 3})

        c = ureg.get_base_units("cm", system=sysname)
        self.assertAlmostEqual(c[0], 0.1283, places=3)
        self.assertEqual(c[1], {"pint": 1.0 / 3})

        c = ureg.get_base_units("inch**2", system=sysname)
        self.assertAlmostEqual(c[0], 0.326 ** 2, places=3)
        self.assertEqual(c[1], {"pint": 2.0 / 3})

        c = ureg.get_base_units("cm**2", system=sysname)
        self.assertAlmostEqual(c[0], 0.1283 ** 2, places=3)
        self.assertEqual(c[1], {"pint": 2.0 / 3})

    def test_get_base_units_relation(self):
        sysname = "mysys4"
        ureg = UnitRegistry()

        g = ureg.get_group("test-imperial")
        g.add_units("inch", "yard", "pint")

        lines = ["@system %s using test-imperial" % sysname, "mph:meter"]

        s = ureg.System.from_lines(lines, ureg.get_base_units)
        ureg._systems[s.name] = s
        # base_factor, destination_units
        c = ureg.get_base_units("inch", system=sysname)
        self.assertAlmostEqual(c[0], 0.056, places=2)
        self.assertEqual(c[1], {"mph": 1, "second": 1})

        c = ureg.get_base_units("kph", system=sysname)
        self.assertAlmostEqual(c[0], 0.6213, places=3)
        self.assertEqual(c[1], {"mph": 1})

    def test_members_nowarning(self):
        ureg = self.ureg
        for name in dir(ureg.sys):
            dir(getattr(ureg.sys, name))
